@@ -38,6 +38,7 @@
 #include "wasm.h"
 #include "pass.h"
 #include "wasm-builder.h"
+#include "ir/flat.h"
 #include "ir/local-graph.h"
 #include "ir/utils.h"
 #include "dataflow/node.h"
@@ -131,7 +132,7 @@ struct UseFinder {
 };
 
 // Generates a trace: all the information to generate a Souper LHS
-// for a specific set_local whose value we want to infer.
+// for a specific local.set whose value we want to infer.
 struct Trace {
   Graph& graph;
   Node* toInfer;
@@ -637,6 +638,7 @@ struct Souperify : public WalkerPass<PostWalker<Souperify>> {
 
   void doWalkFunction(Function* func) {
     std::cout << "\n; function: " << func->name << '\n';
+    Flat::verifyFlatness(func);
     // Build the data-flow IR.
     DataFlow::Graph graph;
     graph.build(func, getModule());
@@ -688,4 +690,3 @@ Pass *createSouperifySingleUsePass() {
 }
 
 } // namespace wasm
-
