@@ -17,6 +17,8 @@
 #ifndef wasm_wasm_type_h
 #define wasm_wasm_type_h
 
+#include "wasm-features.h"
+
 namespace wasm {
 
 enum Type {
@@ -25,6 +27,7 @@ enum Type {
   i64,
   f32,
   f64,
+  v128,
   unreachable // none means no type, e.g. a block can have no return type. but
               // unreachable is different, as it can be "ignored" when doing
               // type checking across branches
@@ -32,11 +35,13 @@ enum Type {
 
 const char* printType(Type type);
 unsigned getTypeSize(Type type);
+FeatureSet getFeatures(Type type);
 Type getType(unsigned size, bool float_);
 Type getReachableType(Type a, Type b);
 bool isConcreteType(Type type);
 bool isFloatType(Type type);
 bool isIntegerType(Type type);
+bool isVectorType(Type type);
 
 } // namespace wasm
 
